import json

from flask import Blueprint, make_response, request, current_app, send_file
from flask_jsonpify import jsonpify
from flask.helpers import NotFound

import demjson3 as demjson

from .controllers import Controllers
from .sources import extract_text_fields
from .logger import logger, logging
from .utils.file_maker import get_csv, get_xls, get_xlsx
from .query import Query


def default_rules(field):
    if field.get('es:title') or field.get('es:hebrew'):
        if field.get('es:keyword'):
            return [('exact', '^10')]
        else:
            return [('inexact', '^3'), ('natural', '.hebrew^10')]
    elif field.get('es:boost'):
        if field.get('es:keyword'):
            return [('exact', '^10')]
        else:
            return [('inexact', '^10')]
    elif field.get('es:keyword'):
        return [('exact', '')]
    else:
        return [('inexact', '')]


class APIESBlueprint(Blueprint):

    def __init__(self, app,
                 sources,
                 es_client,
                 search_indexes,
                 document_index,
                 dont_highlight=None,
                 text_field_rules=default_rules,
                 text_field_select=None,
                 multi_match_type='most_fields',
                 multi_match_operator='and',
                 debug_queries=False,
                 query_cls=Query):
        super().__init__('apies', 'apies')

        if debug_queries:
            logger.setLevel(logging.DEBUG)
        else:
            logger.setLevel(logging.INFO)

        if dont_highlight is not None:
            # deprecation message:
            logger.warning('dont_highlight is deprecated, use request parameters instead')

        self.controllers = Controllers(
            search_indexes=search_indexes,
            text_fields=extract_text_fields(sources, text_field_rules, text_field_select, debug_queries),
            document_index=document_index,
            multi_match_type=multi_match_type,
            multi_match_operator=multi_match_operator,
            debug_queries=debug_queries,
            query_cls=query_cls
        )

        self.add_url_rule(
            '/get/<path:doc_id>',
            'get_document_handler',
            self.get_document_handler,
            methods=['GET']
        )
        self.add_url_rule(
            '/search/count',
            'simple_count_handler',
            self.count_handler,
            methods=['GET']
        )
        self.add_url_rule(
            '/search/<string:types>',
            'dynamic_search_handler',
            self.search_handler,
            methods=['GET']
        )
        self.add_url_rule(
            '/download/<string:types>',
            'download',
            self.download,
            methods=['GET']
        )

        app.config['ES_CLIENT'] = es_client

    def search_handler(self, types):
        es_client = current_app.config['ES_CLIENT']

        try:
            types_formatted = str(types).split(',')
            filters = request.values.get('filter')
            lookup = request.values.get('lookup')
            search_term = request.values.get('q')
            term_context = request.values.get('context')
            extra = request.values.get('extra')
            from_date = request.values.get('from_date')
            to_date = request.values.get('to_date')
            size = request.values.get('size', 10)
            offset = request.values.get('offset', 0)
            order = request.values.get('order')
            highlight = [x.strip() for x in request.values.get('highlight', '').split(',') if x]
            snippets = [x.strip() for x in request.values.get('snippets', '').split(',')]
            result = self.controllers.search(
                es_client, types_formatted, search_term,
                from_date=from_date,
                to_date=to_date,
                size=size,
                offset=offset,
                filters=filters,
                lookup=lookup,
                term_context=term_context,
                extra=extra,
                score_threshold=0, 
                sort_fields=order,
                highlight=highlight,
                snippets=snippets
            )
        except Exception as e:
            logger.exception('Error searching %s for types: %s ' % (search_term, str(types)))
            result = {'error': str(e)}
        return jsonpify(result)

    def download(self, types):
        """
        Performs a search and returns the results in a file (CSV or Excel) response

        :param types: The types of the document
        :return Response: The file response with the results of the query
        """

        # Get values from the config
        es_client = current_app.config['ES_CLIENT']

        # Get parameters from the query string
        try:
            types_formatted = str(types).split(',')
            search_term = request.values.get('q')
            term_context = request.values.get('context')
            extra = request.values.get('extra')
            from_date = request.values.get('from_date')
            to_date = request.values.get('to_date')
            size = request.values.get('size', 10)
            offset = request.values.get('offset', 0)
            filters = request.values.get('filter')
            lookup = request.values.get('lookup')
            order = request.values.get('order')

            # Get the query results
            result = self.controllers.search(es_client,
                                             types_formatted,
                                             search_term,
                                             from_date,
                                             to_date,
                                             size,
                                             offset,
                                             filters,
                                             lookup,
                                             term_context,
                                             extra,
                                             score_threshold=0,
                                             sort_fields=order)

        except Exception as e:
            logging.exception('Error searching %s for types: %s ' % (search_term, str(types)))
            result = {'error': str(e)}

        # Get the file name from the querystring
        file_name = request.values.get('file_name') or 'budgetkey'

        # Do the column mapping
        column_mapping = request.values.get('column_mapping')
        if column_mapping:
            column_mapping = json.loads(column_mapping)

        # Get the file name and format from the query string, or give them default values
        file_format = request.values.get('file_format')
        if file_format == 'csv':
            file = get_csv(result, column_mapping)

            # Make the response object
            response = make_response(file)
            response.headers["Content-Disposition"] = "attachment; filename={}".format(file_name + '.csv')
            response.headers["Content-type"] = 'text/csv'

        elif file_format == 'xls':
            file_stream = get_xls(result, column_mapping)

            # Make the response object
            response = send_file(file_stream,
                                 as_attachment=True,
                                 mimetype='application/vnd.ms-excel',
                                 attachment_filename=file_name + '.xlsx')

        else:
            file_stream = get_xlsx(result, column_mapping)

            # Make the response object
            response = send_file(file_stream,
                                 as_attachment=True,
                                 mimetype='application/vnd.openxmlformats-officedocument.spreadsheetml.sheet',
                                 attachment_filename=file_name + '.xlsx')

        return response

    def count_handler(self):
        es_client = current_app.config['ES_CLIENT']

        config = request.values.get('config')
        try:
            config = demjson.decode(config)
            search_term = request.values.get('q')
            from_date = request.values.get('from_date')
            to_date = request.values.get('to_date')
            term_context = request.values.get('context')
            extra = request.values.get('extra')
            result = self.controllers.count(
                es_client, search_term, from_date, to_date, config, term_context, extra
            )
        except Exception as e:
            logger.exception('Error counting with config %r', config)
            result = {'error': str(e)}
        return jsonpify(result)


    def get_document_handler(self, doc_id):
        es_client = current_app.config['ES_CLIENT']
        type_ = request.values.get('type')

        result = self.controllers.get_document(
            es_client, doc_id, type_
        )
        if result is None:
            logger.warning('Failed to fetch document for %r', doc_id)
            raise NotFound()
        return jsonpify(result)



import csv
import xlwt
import xlsxwriter

from io import BytesIO, StringIO


def get_csv(es_result, column_mapping):
    """
    Creates a string for a csv file through the Python csv writer and StringIo

    :param (dict): The result of the ElasticSearch search
    :param column_mapping (dict): An dict mapping the desired column name to the actual field name in Elasticsearch.
    This can also be a nested field.
    For example:
    {
        "address.city":"עיר",
        "details.budget": "תקציב"
    }

    :return column_mapping (dict): A dict mapping the column names in the original data, to the desired column headers
    for the output file
    """

    # Create a string stream and writer object, passing the stream
    file_stream = StringIO()
    writer = csv.writer(file_stream)

    # Get ordered lists of column headers (the 'titles' of the columns) and the column names (how the fields are called
    # in ElasticSearch
    document_source = es_result['search_results'][0]['source']
    column_headers, column_names = _get_column_headers_and_names(document_source, column_mapping)

    # Write the header row
    writer.writerow(column_headers)

    # Loop over the documents, for each document loop over the fields and retrieve the field from the document
    documents = es_result['search_results']
    for document in documents:
        document_row = []
        for field_name in column_names:
            field_value = _get_field_value(field_name, document['source'])
            document_row.append(field_value)
        # Write the document field
        writer.writerow(document_row)

    # Return the stream string
    return file_stream.getvalue()


def get_xls(es_result, column_mapping):
    """
    Creates a stream with the Excel file, the column headers, and the result rows

    :param es_result: The result of the ElasticSearch search
    :param column_mapping (dict): A dict mapping the column names in the original data, to the desired column headers
    for the output file

    :return (BytesIO): A stream with the Excel file
    """

    # Create a bytes stream
    file_stream = BytesIO()

    # Create an excel workbook and sheet
    workbook = xlwt.Workbook()
    worksheet = workbook.add_sheet('result')

    # Get ordered lists of the column headers ('column titles') and the column names (how the fields are actually called
    # in the ElasticSearch results)
    document_source = es_result['search_results'][0]['source']
    column_headers, column_names = _get_column_headers_and_names(document_source, column_mapping)

    # Write the first row of column names
    for index, column_name in enumerate(column_headers):
        worksheet.write(0, index, column_name)

    # Write the rows
    documents = es_result['search_results']
    # Loop over the documents
    for document_index, document in enumerate(documents):
        # Loop over the columns
        for column_index, column_name in enumerate(column_names):
            column_value = _get_field_value(column_name, document['source'])
            worksheet.write(document_index+1, column_index, column_value)

    # Save the workbook to the stream
    workbook.save(file_stream)
    file_stream.seek(0)

    return file_stream


def get_xlsx(es_result, column_mapping):
    """
    Creates a stream with the Excel file, the column headers, and the result rows

    :param es_result (dict): The result of the Elasticsearch search
    :param column_mapping (dict): A dict mapping the column names in the original data, to the desired column headers
    for the output file

    :return (IO.BytesIO): A stream with the Excel file
    """

    # Create a file stream
    output = BytesIO()

    # Create a workbook with the file stream, and worksheet object
    workbook = xlsxwriter.Workbook(output)
    worksheet = workbook.add_worksheet()

    # Get ordered lists of the column headers ('column titles') and the column names (how the fields are actually called
    # in the ElasticSearch results)
    document_source = es_result['search_results'][0]['source']
    column_headers, column_names = _get_column_headers_and_names(document_source, column_mapping)

    # Write the first row of column names
    for index, column_header in enumerate(column_headers):
        worksheet.write(0, index, column_header)

    # Write the rows
    documents = es_result['search_results']
    for document_index, document in enumerate(documents):
        # Loop over the columns
        for column_index, column_name in enumerate(column_names):
            column_value = _get_field_value(column_name, document['source'])
            worksheet.write(document_index+1, column_index, column_value)

    # Close the workbook before streaming the data.
    workbook.close()

    # Rewind the buffer.
    output.seek(0)

    return output


def _get_column_headers_and_names(document_source, column_mapping):
    """
    Creates ordered lists of column headers (the 'titles of the columns), and column names (the name of the fields in
    the ElasticSearch results, given the column mapping.

    So, for example, given the following column_mapping:
    {
      "עיר": "address.city"
      "תקציב": "details.budget"
    }
    The returned values will be:
    - column_headers: ['עיר', 'תקציב']
    - column_names: ['address.city', 'details.budget']

    If there is no column_mapping sent, the column_headers will be the same as the column_names

    :param document_source: A single document from the Elasticsearch search result
    :param column_mapping (dict): A dict mapping the column names in the original data, to the desired column headers
    for the output file (see example above)

    :return (tuple): A tuple consisting of column_headers and column_names, both lists (see example above)
    """

    # In case of column mapping, set the order of the columns and then get the headers list and the names list
    if column_mapping:
        ordered_columns = [(field, column_mapping[field]) for field in column_mapping]
        column_headers = [(field[1]) for field in ordered_columns]
        column_names = [(field[0]) for field in ordered_columns]

    # If no column_mapping is sent, simply set the order, and the column_names are the same as the column_headers
    else:
        column_headers = [field for field in document_source]
        column_names = column_headers

    return column_headers, column_names


def _get_field_value(field_to_get, object_to_search):
    """
    Gets the value of a nested object, for example, if we search
    'address.city.neighborhood' in the object:
    {'address':
        {'city':
            {'neighborhood': 'shapira'}
        }
    }
    The returned value will be 'shapira'

    :param field_to_get:
    :param object_to_search:
    :return (string): The value of the field that needs to be searched
    """

    # Split the field to get up in parts (in the example above: ['address', 'city', 'neighborhood']
    field_parts = field_to_get.split('.')

    # Keep popping field parts and retrieving the nested objects until the last part is reached.
    while len(field_parts) > 0:
        # Get the first 'field part' (in the example above, 'address') and pop it from the parts
        field_part = field_parts.pop(0)
        # Get the nested object from the object (in the example above: {'address': {'...':{'...': {}}}}
        object_to_search = object_to_search.get(field_part, '')

    return object_to_search



from copy import copy

from datapackage import Package, Resource
from tableschema import Field

from .logger import logger


def _process_field(field: Field, rules, field_select, ret, prefix):
    schema_type = field['type']
    if schema_type == 'array':
        field = copy(field)
        field['type'] = field['es:itemType']
        return _process_field(field, rules, field_select, ret, prefix)
    enabled = field.get('es:index', True) and not field.get('es:exclude', False)
    subschema = {'fields': []}
    if enabled:
        if schema_type == 'object':
            subschema = field['es:schema']
            _process_schema(subschema, rules, field_select, ret, prefix + field['name'] + '.')
        elif schema_type == 'string':
            search_field = prefix + field['name']
            if field_select is None or search_field in field_select:
                for kind, suffix in rules(field):
                    ret.append((kind, search_field + suffix))


def _process_schema(schema, rules, field_select, ret=[], prefix=''):
    fields = schema['fields']
    for f in fields:
        _process_field(f, rules, field_select, ret, prefix)
    return ret


def extract_text_fields(sources, text_field_rules, text_field_select, debug=False):

    sources = [src if isinstance(src, Package) else Package(src)
               for src in sources]

    ret = {}
    source: Package
    for source in sources:
        resource: Resource = source.resources[0]
        type_name = resource.name
        type_text_field_select = text_field_select.get(type_name) if text_field_select else None
        schema = resource.schema.descriptor
        text_fields = _process_schema(schema, text_field_rules, type_text_field_select, ret=[])
        ret[type_name] = text_fields
        if debug:
            logger.info('TEXT FIELDS (for %s):\n%s', type_name, ', '.join(map(str, text_fields)))


    return ret



from .logger import logger
from .query import Query

import elasticsearch

class Controllers():

    def __init__(self,
                 search_indexes,
                 text_fields,
                 document_index,
                 multi_match_type='most_fields',
                 multi_match_operator='and',
                 debug_queries=False,
                 query_cls=Query):

        self.text_fields = text_fields
        self.search_indexes = search_indexes
        self.document_index = document_index
        self.multi_match_type = multi_match_type
        self.multi_match_operator = multi_match_operator
        self.debug_queries = debug_queries
        self.query_cls = query_cls

    # REPLACEMENTS
    def _do_replacements(self, value, replacements):
        if value is None:
            return None

        if isinstance(value, str):
            for src, dst in replacements:
                value = value.replace(src, dst)
            value = value.replace('</em> <em>', ' ')
            return value

        if isinstance(value, (int, bool, float)):
            return value

        if isinstance(value, list):
            return [self._do_replacements(v, replacements) for v in value]

        if isinstance(value, dict):
            return dict((k, self._do_replacements(v, replacements)) for k, v in value.items())

        assert False, 'Unknown type %r' % value

    def _merge_highlight_into_source(self, source, highlights, highlight, snippets):
        _snippets = source.setdefault('_snippets', dict())
        _highlights = source.setdefault('_highlights', dict())
        for field, highlighted in highlights.items():
            if field in snippets:
                _snippets[field] = highlighted
            elif field in highlight:
                field_parts = field.split('.')
                out_field = []
                src = source
                while len(field_parts) > 0:
                    field = field_parts.pop(0)
                    out_field.append(field)
                    if isinstance(src.get(field), dict):
                        src = src[field]
                    else:
                        break
                out_field = '.'.join(out_field)
                _highlights[out_field] = highlighted[0]
        return source

    # UTILS
    def _validate_types(self, types):
        if 'all' in types:
            types = self.search_indexes

        for type_name in types:
            if type_name not in self.search_indexes:
                raise ValueError('not a real type %s' % type_name)
        return dict((k, v) for k, v in self.search_indexes.items() if k in types)

    # Main API
    def search(self,
               es_client,
               types,
               term,
               *,
               from_date=None,
               to_date=None,
               size=10,
               offset=0,
               filters=None,
               lookup=None,
               term_context=None,
               extra=None,
               score_threshold=0,
               sort_fields=None,
               highlight=None,
               snippets=None):
        search_indexes = self._validate_types(types)

        query = self.query_cls(search_indexes)
        if term:
            query = query.apply_term(
                term, self.text_fields,
                multi_match_type=self.multi_match_type,
                multi_match_operator=self.multi_match_operator)

        if term_context:
            query = query.apply_term_context(term_context, self.text_fields)

        # Apply the filters
        query = query.apply_filters(filters)

        # Apply the lookup
        query = query.apply_lookup(lookup)

        # Apply extra processing
        query = query.apply_extra(extra)

        # Apply sorting - if there are fields to sort by, apply the scoring as the sorting
        if sort_fields is None:
            if term:
                query.apply_scoring()
            else:
                query.apply_sorting({'score': {'order': 'desc'}}, 0)
        else:
            query.apply_sorting(sort_fields, score_threshold)

        # Apply pagination
        query = query.apply_pagination(size, offset)

        # Apply highlighting
        if term and highlight or snippets:
            query = query.apply_highlighting(term, highlight, snippets)

        # Ensure correct counts
        query = query.apply_exact_total()

        # Apply the time range
        query = query.apply_time_range(from_date, to_date)

        # Execute the query
        query_results = query.run(es_client, self.debug_queries)
        query_results = query_results['responses']
        hits = []
        total_overall = 0
        search_counts = dict()
        for _type, result in zip(query.types, query_results):
            result_hits = result.get('hits', {})
            for i, hit in enumerate(result_hits.get('hits', [])):
                hit['_type'] = _type
                hits.append((i, hit))
            count = result_hits.get('total', {}).get('value', 0)
            total_overall += count
            search_counts[_type] = dict(total_overall=count)
            if 'hits' not in result or 'hits' not in result['hits']:
                logger.warning('no hits element for query for type %s: %r', _type, result)
        hits = [j[1] for j in sorted(hits, key=lambda i: i[0])]

        default_sort_score = (0,)
        search_results = [
            dict(
                source=self._merge_highlight_into_source(
                    hit['_source'],
                    hit['highlight'],
                    highlight,
                    snippets
                ),
                type=hit['_type'],
                score=hit['_score'] or hit.get('sort', default_sort_score)[0]
            ) if 'highlight' in hit else dict(
                source=hit['_source'],
                type=hit['_type'],
                score=hit['_score'] or hit.get('sort', default_sort_score)[0]
            )
            for hit in hits
        ]

        search_counts['_current'] = dict(
            total_overall=total_overall
        )
        return dict(
            search_counts=search_counts,
            search_results=search_results
        )

    def count(self, es_client, term, from_date, to_date, config, term_context, extra):
        counts = {}
        for item in config:
            doc_types = item['doc_types']
            search_indexes = self._validate_types(doc_types)
            filters = item['filters']
            id = item['id']
            query_results = self.query_cls(search_indexes)
            if term:
                query_results = query_results.apply_term(
                    term, self.text_fields,
                    multi_match_type=self.multi_match_type,
                    multi_match_operator=self.multi_match_operator
                )
            if term_context:
                query_results = query_results.apply_term_context(term_context, self.text_fields)

            query_results = query_results\
                .apply_filters(filters)\
                .apply_pagination(0, 0)\
                .apply_time_range(from_date, to_date)\
                .apply_exact_total()

            # Apply extra processing
            if extra:
                query_results = query_results.apply_extra(extra)

            query_results = query_results.run(es_client, self.debug_queries)
            counts[id] = dict(
                total_overall=sum(
                    results['hits']['total']['value']
                    for results in query_results['responses']
                )
            )
        return dict(
            search_counts=counts
        )


    def get_document(self, es_client, doc_id, doc_type=None):
        try:
            index = self.document_index
            if doc_type is not None:
                types = [doc_type]
                types = self._validate_types(types)
                index = types[doc_type]
            logger.debug('FETCH %r in %s (%r)', doc_id, index, doc_type)
            result = es_client.get(index, doc_id)
            return result.get('_source')
        except elasticsearch.exceptions.NotFoundError:
            return None



import demjson3 as demjson
import json
from elasticsearch import Elasticsearch

from .logger import logger


# ### QUERY DSL HANDLING
class Query():

    def __init__(self, search_indexes):
        self.types = list(search_indexes.keys())
        self.filtered_type_names = set(self.types)
        self.indexes = list(search_indexes.values())
        self.q = dict((t, {}) for t in self.types)

    def __str__(self):
        return demjson.encode(self.q)

    def run(self, es_client: Elasticsearch, debug):
        if debug:
            logger.debug('QUERY (for %s):\n%s', self.types[0],
                         json.dumps(self.q[self.types[0]], indent=2, ensure_ascii=False))
            if len(self.types) > 1:
                logger.debug('QUERY (for %s):\n%s', self.types[-1],
                            json.dumps(self.q[self.types[-1]], indent=2, ensure_ascii=False))

        body = ''.join(
            '{}\n{}\n'.format(
                json.dumps(dict(index=index)),
                json.dumps(self.q[t])
            )
            for t, index in zip(self.types, self.indexes)
            if t in self.filtered_type_names
        )
        return es_client.msearch(body)

    def query_bool(self, t):
        return self.q[t].setdefault('query', {})\
                        .setdefault('function_score', {})\
                        .setdefault('query', {})\
                        .setdefault('bool', {})

    def must(self, t):
        return self.query_bool(t)\
                        .setdefault('must', [])

    def should(self, t):
        return self.query_bool(t)\
                        .setdefault('should', [])

    def filter(self, t):
        return self.query_bool(t)\
                        .setdefault('filter', {})\
                        .setdefault('bool', {})

    def must_not(self, t):
        return self.query_bool(t)\
                        .setdefault('must_not', [])

    def apply_term(self, term, text_fields,
                   multi_match_type='most_fields', multi_match_operator='and'):
        for type_name in self.types:
            search_fields = text_fields[type_name]
            search_fields = dict(
                (k, [x[1] for x in search_fields if x[0] == k])
                for k in ('exact', 'inexact', 'natural')
            )
            matchers = []

            # Multimatch for inexact fields
            matchers.append(dict(
                multi_match=dict(
                    query=term,
                    fields=[f for f in search_fields['inexact'] + search_fields['natural']],
                    type=multi_match_type,
                    operator=multi_match_operator,
                    tie_breaker=0.3
                )
            ))

            # Tuples
            parts = term.split()
            parts = [term] + parts + [' '.join(z) for z in zip(parts[:-1], parts[1:])]
            for field in search_fields['exact']:
                fparts = field.split('^')
                if len(fparts) == 1:
                    matchers.append(dict(
                        terms={
                            field: tuple(set(parts))
                        }
                    ))
                else:
                    matchers.append(dict(
                        terms={
                            fparts[0]: tuple(set(parts)),
                            'boost': float(fparts[1])
                        }
                    ))

            # Apply boosters
            if len(matchers) > 0:
                self.must(type_name).append(dict(
                    bool=dict(
                        should=matchers,
                        minimum_should_match=1
                    )
                ))

        return self

    def apply_term_context(self, terms, text_fields):
        multi_match_type = 'most_fields'
        multi_match_operator = 'or'
        for type_name in self.types:
            search_fields = text_fields[type_name]
            search_fields = [
                x[1] for x in search_fields if x[0] == 'inexact'
            ]
            matcher = dict(
                multi_match=dict(
                    query=terms,
                    fields=search_fields,
                    type=multi_match_type,
                    operator=multi_match_operator,
                    tie_breaker=0.3
                )
            )
            self.filter(type_name).setdefault('must', []).append(matcher)

        return self

    def apply_scoring(self):
        for type_name in self.types:
            fs = self.q[type_name].setdefault('query', {}).setdefault('function_score', {})
            fs.update(dict(
                boost_mode='multiply',
                field_value_factor=dict(
                    field='score',
                    modifier='sqrt',
                    missing=1
                )
            ))
        return self

    def apply_sorting(self, sort_fields, score_threshold):
        # Apply the scoring threshold - since we are no longer sorting by score, it is important to use a score
        # threshold so as not to get irrelevant results
        for type_name in self.types:
            q = self.q[type_name]
            q.setdefault('min_score', score_threshold)
            if isinstance(sort_fields, str):
                if sort_fields[0] == '-':
                    trimmed_fields = sort_fields[1:]
                    sort = {trimmed_fields: {'order': 'desc'}}
                else:
                    sort = {sort_fields: {'order': 'asc'}}
            else:
                sort = sort_fields

            # Then sort by the sort fields, for example - {'__last_modified_at': {'order': 'desc'}}
            q.setdefault('sort', sort)

        return self

    def apply_pagination(self, page_size, offset):
        for type_name in self.types:
            self.q[type_name].update({
                'size': int(page_size),
                'from': int(offset)
            })
        return self

    def apply_highlighting(self, term, highlight, snippets):
        for type_name in self.types:
            self.q[type_name]['highlight'] = dict(
                fields=dict(
                    (f, dict() if f in snippets else dict(number_of_fragments=0)) for f in highlight + snippets
                ),
                highlight_query=dict(
                    multi_match=dict(
                        query=term,
                        fields=highlight + snippets,
                    )
                )
            )
        return self

    def parse_filter_op(self, k, v):
        must = True
        parts = k.split('__')
        op = None

        if len(parts) > 1 and parts[-1] in ('gt', 'gte', 'lt', 'lte', 'eq', 'not', 'like', 'bounded'):
            op = parts[-1]
            k = '__'.join(parts[:-1])

        if op == 'not':
            must = False
            op = None

        if op is not None:
            if op == 'like':
                ret = dict(
                    match={
                        k: v
                    }
                )
            elif op == 'bounded':
                ret = dict(
                    geo_bounding_box={
                        k: {
                            'top_left': dict(
                                lat=v[0][1],
                                lon=v[0][0],
                            ),
                            'bottom_right': dict(
                                lat=v[1][1],
                                lon=v[1][0],
                            )
                        }
                    }
                )
            else:
                ret = dict(
                    range={
                        k: {
                            op: v
                        }
                    }
                )
        else:
            if isinstance(v, list):
                ret = dict(
                    terms={
                        k: v
                    }
                )
            else:
                ret = dict(
                    term={
                        k: v
                    }
                )
        return ret, must

    def _process_complex(self, param):
        if not param:
            return None

        if isinstance(param, str):
            if param.startswith('[') and param.endswith(']'):
                pass
            elif not param.startswith('{'):
                param = '{' + param + '}'
            param = demjson.decode(param)

        if isinstance(param, dict):
            param = [param]

        if isinstance(param, list):
            should_clauses = dict()
            for i in param:
                bool_clause = {}
                type_names = self.types
                for k, v in i.items():
                    if k == '_type':
                        if not isinstance(v, list):
                            type_names = [v]
                        else:
                            type_names = v
                    else:
                        clause, positive = self.parse_filter_op(k, v)
                        if positive:
                            bool_clause.setdefault('must', []).append(clause)
                        else:
                            bool_clause.setdefault('must_not', []).append(clause)
                for type_name in type_names:
                    should_clauses.setdefault(type_name, []).append(dict(bool=bool_clause))
            return should_clauses

        return None

    def apply_filters(self, filters):
        should_clauses = self._process_complex(filters)
        if not should_clauses:
            return self

        for type_name, clauses in should_clauses.items():
            if len(clauses) > 0:
                flt = self.filter(type_name)
                flt.setdefault('should', []).extend(clauses)
                flt['minimum_should_match'] = 1
        self.filtered_type_names = set(should_clauses.keys())
        return self

    def apply_lookup(self, lookup):
        should_clauses = self._process_complex(lookup)
        if not should_clauses:
            return self

        for type_name, clause in should_clauses.items():
            self.should(type_name).append(
                dict(
                    bool=dict(
                        should=clause
                    )
                )
            )
            self.query_bool(type_name)['minimum_should_match'] = 1
        self.filtered_type_names = set(should_clauses.keys())
        return self

    def apply_time_range(self, from_date, to_date):
        if None not in (from_date, to_date):
            for type_name in self.types:
                self.must(type_name).extend([
                    dict(
                        range=dict(
                            __date_range_from=dict(
                                lte=to_date
                            )
                        )
                    ),
                    dict(
                        range=dict(
                            __date_range_to=dict(
                                gte=from_date
                            )
                        )
                    ),
                ])
        return self

    def apply_exact_total(self):
        for type_name in self.types:
            self.q[type_name]['track_total_hits'] = True
        return self

    def apply_month_aggregates(self):
        for type_name in self.types:
            self.q[type_name].setdefault('aggs', {}).update(dict(
                timeline=dict(
                    terms=dict(
                        field='__date_range_months',
                        size=2500,
                        order=dict(_term='asc')
                    )
                )
            ))
        return self

    def apply_extra(self, extras):
        return self